/* ©  Copyright Steve Simon, 2013 */
#include <u.h>
#include <libc.h>
#include <fcall.h>
#include <thread.h>
#include <9p.h>
#include <pool.h>
#include "httpfs.h"

typedef struct Aux Aux;
struct Aux {
	Node *node;

	Growbuf *gb;	/* snoop on reads and parse them if they are html */
	vlong off;
	int eof;
};

int Dumptree;
int Dumpreq;
int Dumpnodes;
int Dumpparse;

Sess *Session;

static void
dumptree(Node *n, int in)
{

	for (; n; n = n->next){
		print("%*s %D\n", in, "", &n->dir);
		if (n->child)
			dumptree(n->child, in+2);
	}
}

static void
dircp(Dir *d, Dir *s)
{
	*d = *s;
	d->uid = estrdup9p(d->uid);
	d->gid = estrdup9p(d->gid);
	d->muid = estrdup9p(d->muid);
	d->name = estrdup9p(d->name);
}

static vlong
hash(char *s)
{
	vlong h;
	uchar *p;

	h = 0;
	for(p = (uchar*)s; *p; p++)
		h = h*37 + *p;
	return h;
}

static char *
getpath(Node *n, long len)
{
	long sz;
	char *buf;

	if(n->parent){
		sz = len + strlen(n->dir.name) +1;		/* +1 for '/' */
		buf = getpath(n->parent, sz);
		strcat(buf, n->dir.name);
		if(len)
			strcat(buf, "/");
	}
	else
		buf = emalloc9p(len + 1);			/* +1 for '\0' */
	return buf;
}

		
static Dir
protodir(Dir *parent, char *name, vlong qpath, int mode, vlong length, 
	char *uid, char *gid, char *muid, long mtime, long atime)
{
	Dir d;
	long now = time(nil);

	if(uid == nil){
		if(parent)
			uid = parent->uid;
		if(uid == nil)
			uid = "mire";
	}
	
	if(gid == nil){
		if(parent)
			gid = parent->gid;
		if(gid == nil)
			gid = "mire";
	}

	if(muid == nil){
		if(parent)
			muid = parent->muid;
		if(muid == nil)
			muid = uid;
	}

	/*
	 * we set the default mtime to 1/1/70 so it is
	 * obvious the file has not been stat()'ed yet
	 */
	if(mtime == 0){
		if(parent)
			mtime = parent->mtime;
		if(mtime == -1L)
			mtime = 0;
	}

	if(atime == 0){
		if(parent)
			atime = parent->atime;
		if(atime == -1L)
			atime = now;
	}

	/* cannot inherit or default these */
	d.name = estrdup9p(name);
	d.length = length;
	d.mode = mode;

	d.qid.type = (mode & DMDIR)? QTDIR: QTFILE;
	d.qid.path = qpath;
	d.qid.vers = now;
	d.uid = estrdup9p(uid);
	d.gid = estrdup9p(gid);
	d.muid = estrdup9p(muid);
	d.mtime = mtime;
	d.atime = atime;

	return d;
}

static Node *
mknode(Node **np, char *name, char *path, int mode, long length, long mtime)
{
	Node *t;
	Dir *proto;

	for(t = *np; t; t = t->next)
		if(strcmp(t->dir.name, name) == 0)
			break;
	if(t)
		return t;

	proto = nil;
	if(*np)
		proto = &(*np)->dir;

	if(Dumpnodes)
		print("add %s\n", path);

	t = emalloc9p(sizeof(Node));
	t->dir = protodir(proto, name, hash(path), mode,
		length, nil, nil, nil, mtime, 0);
	t->next = *np;
	t->path = path;
	*np = t;
	return t;
}

static int
isfile(char *name)
{
	int i, l;
	char *p;
	char *Exts[] = {
		".htm",	".html",	".aspx",
		".pdf",	".txt",	".ps",
		".jpg",	".jpeg",	".gif",	".png",
		".man",	".c",	".h",	".cpp",	".C",	".hpp",	".s",	".asm",
		".xml",	".json",	".dtd",
		".wav", ".mp3", ".aac"
		".mov",	".h264", ".avi", ".qt",
		".gz",	".bz2",	".Z",	".z",
		".zip",	".tar",	".tbz",	".tgz",	".bun",	".shar",
		".ms", ".nr", ".tbl", ".pic", ".eqn", ".tex",
		".doc", ".xls", ".ppt", ".docx", ".xlsx", ".pptx",
		".exe",
	};

	if((p = strrchr(name, '.')) == nil)
		return 0;
	if(p[1] == 0)
		return 0;
	if(strcmp(p, ".dir") == 0)
		return 0;
	for(i = 0 ; i < nelem(Exts); i++){
		if(cistrcmp(Exts[i], p) == 0)
			return 1;
		l = strlen(Exts[i]);
		if(cistrncmp(Exts[i], p, l) == 0 && p[l] == '?')		/* UTF-8 breaks this */
			return 1;	
	}

	fprint(2, "'%s' - file has unknown extension\n", name);
	return 1;
}

void
newpath(Sess *s, char *path, int mode, long mtime, vlong length)
{
	long now;
	char *p, *name;
	Node *c, **np;

	cleanname(path);

	name = path;
	while(*name == '/')
		name++;

	now = time(nil);
	np = &s->root->child;
	while((p = strchr(name, '/')) != nil){
		*p = 0;
		c = mknode(np, name, estrdup9p(path), 0755|DMDIR, 0, now);
		c->dir.mode |= 0755;
		np = &c->child;
		*p = '/';
		name = p+1;
	}

	if(*name){
		if(! isfile(name))
			mode |= DMDIR;
		mknode(np, name, path, mode, length, mtime);
	}
}

static void
responderrstr(Req *r)
{
	char e[ERRMAX];

	*e = 0;
	rerrstr(e, sizeof(e));
	respond(r, e);
}

static void
fsattach(Req *r)
{
	Aux *a;
	char *spec = r->ifcall.aname;

	if(spec && *spec){
		respond(r, "invalid attach specifier");
		return;
	}

	a = emalloc9p(sizeof(Aux));
	r->fid->aux = a;
	a->node = Session->root;
	
	r->ofcall.qid = a->node->dir.qid;
	r->fid->qid = r->ofcall.qid;
	respond(r, nil);
}

static char*
fsclone(Fid *fid, Fid *nfid)
{
	Aux *a = fid->aux;
	Aux *na;

	na = emalloc9p(sizeof(Aux));
	na->node = a->node;
	nfid->aux = na;
	return nil;
}


static int
dirgen(int slot, Dir *d, void *aux)
{
	Node *n;
	Aux *a = aux;

	for(n = a->node->child; n; n = n->next)
		if(--slot < 0)
			break;
	if(n == nil)
		return -1;
	dircp(d, &n->dir);
	return 0;
}

static void
refresh(Node *n)
{
	if((n->dir.mode & DMDIR) != DMDIR)
		return;
	if(n->confirmed)
		return;

	switch(parse(Session, n->path)){
	case -1:			/* not found */
		n->confirmed = 0;
		break;
	case 0:				/* found but no html */
		n->confirmed = 0;
		n->dir.mode |= ~DMDIR;
		n->dir.qid.type |= ~QTDIR;
		break;
	case 1:				/* is html */
		n->confirmed = 0;
		break;
	default:
		abort();
	}
}

static char*
fswalk1(Fid *fid, char *name, Qid *qid)
{
	Node *n;
	Aux *a = fid->aux;

	if(strcmp(name, "..") == 0){
		a->node = a->node->parent;
		fid->qid = a->node->dir.qid;
		*qid = a->node->dir.qid;
		return nil;
	}

	for(n = a->node->child; n; n = n->next){
		n->parent = a->node;
		if(strcmp(n->dir.name, name) == 0){
			refresh(n);
			fid->qid = n->dir.qid;
			*qid = n->dir.qid;
			a->node = n;
			return nil;
		}
	}

	return "not found";
}

static void
fsstat(Req *r)
{
	Aux *a = r->fid->aux;

	if((r->fid->qid.type & QTDIR) == 0 && !a->node->confirmed){
		if(ht_head(Session, a->node->path, &a->node->dir.length,
				&a->node->dir.mtime) < 0){
			responderrstr(r);
			return;
		}
		a->node->dir.mode = 0644;
		a->node->confirmed = 1;
	}
	dircp(&r->d, &a->node->dir);
	respond(r, nil);
}

static void
fscreate(Req *r)
{
	Aux *a = r->fid->aux;
	char *path;
	int mode;

	if((r->ifcall.perm & DMDIR) == DMDIR)
		mode = 0755|DMDIR;
	else
		mode = 0644;
	path = smprint("%s/%s", a->node->path, r->ifcall.name);
	mknode(&a->node->child, r->ifcall.name, path, mode, 0, time(nil));
	respond(r, nil);
}

static int
is_html(char *path)
{
	char *p;

	if((p = strrchr(path, '.')) == nil)
		return 0;
	if(strcmp(p, ".htm") == 0)
		return 1;
	if(strcmp(p, ".html") == 0)
		return 1;
	if(strcmp(p, ".aspx") == 0)
		return 1;
	return 0;
}


static void
fsopen(Req *r)
{
	Aux *a = r->fid->aux;

	if (r->fid->qid.type & QTDIR){	/* open dir is a noop */
		respond(r, nil);
		return;
	}
	if(ht_head(Session, a->node->path, &a->node->dir.length,
			&a->node->dir.mtime) < 0){
		responderrstr(r);
		return;
	}
	a->node->dir.mode = 0644;
	a->node->confirmed = 1;

	if(is_html(a->node->path))
		a->gb = allocbuf();
	else
		a->gb = nil;

	respond(r, nil);
}

static void
fsread(Req *r)
{
	Aux *a = r->fid->aux;
	long n;

	if (r->fid->qid.type & QTDIR){
		dirread9p(r, dirgen, a);
		respond(r, nil);
		return;
	}

	if((n = ht_get(Session, a->node->path, r->ofcall.data,
			r->ifcall.count, r->ifcall.offset)) < 0){
		responderrstr(r);
		return;
	}

	/* only snoop on sequential reads */
	if(a->gb && a->off == r->ifcall.offset){
		growbuf(a->gb, r->ofcall.data, n);
		a->off += n;
		if(n == 0)
			a->eof = 1;
	}

	r->ofcall.count	= n;
	if(a->node->dir.length < r->ifcall.offset + r->ofcall.count)
		a->node->dir.length = r->ifcall.offset + r->ofcall.count;
	respond(r, nil);
}

static void
fswrite(Req *r)
{
	Aux *a = r->fid->aux;
	long n;

	if((n = ht_put(Session, a->node->path, r->ifcall.data,
			r->ifcall.count, r->ifcall.offset)) < 0){
		responderrstr(r);
		return;
	}

	r->ofcall.count	= n;
	if(a->node->dir.length < r->ifcall.offset + r->ofcall.count)
		a->node->dir.length = r->ifcall.offset + r->ofcall.count;
	respond(r, nil);
}

static void
fsremove(Req *r)
{
	Aux *a;

	a = r->fid->aux;
	if(ht_delete(Session, a->node->path) < 0){
		responderrstr(r);
		return;
	}
	respond(r, nil);
}

static void
fsdestroyfid(Fid *f)
{
	Aux *a = f->aux;

	if(a == nil)
		return;

	if(a->gb && a->eof){
		growbuf(a->gb, "", 1);
		if(parsebuf(Session, a->node->path, a->gb) < 0)
			fprint(2, "parse %s failed - %r\n", a->node->path);
	}
	if(a->gb)
		freebuf(a->gb);
	free(a);
}

	
Srv fs =
{
	.attach=	fsattach,
	.open=		fsopen,
	.create=	fscreate,
	.read=		fsread,
	.write=		fswrite,
	.stat=		fsstat,
	.clone= 	fsclone,
	.walk1= 	fswalk1,
	.destroyfid= fsdestroyfid,
};


static void
usage(void)
{
	fprint(2, "usage: %s url\n", argv0);
	exits("usage");
}

void
main(int argc, char *argv[])
{
	Sess sess;
	char *p, *mtpt, *svs;

	quotefmtinstall();
	fmtinstall('F', fcallfmt);
	fmtinstall('D', dirfmt);

	memset(&sess, 0, sizeof(Sess));

	svs = nil;
	mtpt = nil;
	sess.webfs = "/mnt/web";
	ARGBEGIN{
	case 'D':
		p = EARGF(usage());
		if(strstr(p, "9p"))
			chatty9p++;
		else
		if(strstr(p, "req"))
			Dumpreq++;
		else
		if(strstr(p, "tree"))
			Dumptree++;
		else
		if(strstr(p, "nodes"))
			Dumpnodes++;
		else
		if(strstr(p, "parse"))
			Dumpparse++;
		else
			usage();
		break;
	case 'w':
		sess.webfs = EARGF(usage());
		break;
	case 'm':
		mtpt = EARGF(usage());
		break;
	case 's':
		svs = EARGF(usage());
		break;
	case 'M':
		mainmem->flags |= POOL_NOREUSE|POOL_PARANOIA|POOL_ANTAGONISM;
		break;
	default:
		usage();
	}ARGEND;

	if(argc != 1)
		usage();
	sess.url = argv[0];
	if(strncmp(sess.url, "http://", 7) != 0 &&
	   strncmp(sess.url, "https://", 8) != 0)
		sysfatal("%s - no reccognised URL scheme\n", sess.url);

	if(ht_session(&sess) < 0)
		sysfatal("cannot get session to webfs\n");

	mknode(&sess.root, "/", estrdup9p("/"), 0755|DMDIR, 0, time(nil));

	parse(&sess, "");

	if(Dumptree)
		dumptree(sess.root, 0);

	if(!mtpt && !svs)
		mtpt = "/n/http";

	Session = &sess;
	postmountsrv(&fs, svs, mtpt, MREPL|MCREATE);
	exits(0);
}
