#include <u.h>
#include <libc.h>
#include <draw.h>
#include <html.h>
#include <fcall.h>
#include <thread.h>
#include <9p.h>
#include "httpfs.h"

static struct {
	Rune *str;
	int len;
} Pfx[] = {
	{ L"http://",		7 },
	{ L"https://",		8 },
	{ L"#",				1 },
	{ L"mailto:",		7 },
	{ L"ftp://",		6 },
	{ L"irc://",		6 },
	{ L"file://",		7 },
	{ L"gopher://",		9 },
	{ L"news://",		7 },
	{ L"newsrc://",		9 },
	{ L"nntp://",		7 },
	{ L"telnet://",		9 },
	{ L"wais://",		7 },
	{ L"x-exec://",		9 },	// execute program
	{ L"mid://",		6 },	// message ID for EMail
	{ L"file://",		7 },
	{ L"cid://",		6 },	// content identifiers
	{ L"afs://",		6 },	// andrew FS
};


/* these two are needed by parsehtml */
void *
emalloc(ulong n)
{
	return emalloc9p(n);
}

void *
erealloc(void *p, ulong n)
{
	return erealloc9p(p, n);
}

static int
charset(char *s)
{
	char *meta, *emeta, *charset;

	meta = cistrstr(s, "<meta");
	if(meta == nil)
		return ISO_8859_1;
	for(emeta=meta; *emeta!='>' && *emeta!='\0'; emeta++)
		continue;
	charset = cistrstr(s, "charset=");
	if(charset == nil)
		return ISO_8859_1;
	charset += 8;
	if(*charset == '"')
		charset++;
	if(cistrncmp(charset, "utf-8", 5) || cistrncmp(charset, "utf8", 4))
		return UTF_8;
	return ISO_8859_1;
}

int
resolve(Sess *s, Rune *base, Rune *href, char *path)
{
	Rune *r;
	int i, hlen;

	if(Dumpparse)
		fprint(2, "refresh base='%S' path='%s' href='%S' \n", base, path, href);
	/*
	 * get the length of the method, hostname,
	 * username, password, port etc prefix
	 */
	r = nil;
	if(runestrncmp(base, L"http://", 7) == 0)
		r = runestrchr(base + 7, '/');
	else
	if(runestrncmp(base, L"https://", 8) == 0)
		r = runestrchr(base + 8, '/');

	if(r == nil){
		fprint(2, "resolve: %S - href contains no '/'\n", base);
		return -1;
	}
	hlen = r-base;

	/*
	 * weird scheme-less absolure urls,
	 * see: hget http://plan9.bell-labs.com | grep 'Plan 9'
	 */
	if(runestrncmp(href, L"//", 2) == 0){
		if((r = runestrchr(href + 2, L'/')) == nil){
			fprint(2, "resolve: %S - href contains no '/'\n", base);
			return -1;
		}
		hlen = r-href;

		if((r = runestrchr(base, L':')) == nil){
			fprint(2, "resolve: %S - base contains no ':'\n", base);
			return -1;
		}

		if(runestrncmp(href, r+1, hlen) == 0){
			if(Dumpparse)
				fprint(2, "	odd %S\n", href+hlen);
			newpath(s, smprint("%S", href + hlen), 0, 0, 0);
			return 0;
		}
	}

	/* absolute URL, and it refers to our host */
	if(runestrncmp(href, base, hlen) == 0){
		if(Dumpparse)
			fprint(2, "	abs %S\n", href+hlen);
		newpath(s, smprint("%S", href + hlen), 0, 0, 0);
		return 0;
	}

	/* absolute URL, but it points elsewhere */
	for(i = 0; i < nelem(Pfx); i++)
		if(runestrncmp(href, Pfx[i].str, Pfx[i].len) == 0)
			break;
	if(i < nelem(Pfx))
		return -1;

	/* absolute path but on the local host */
	if(*href == L'/'){
		if(Dumpparse)
			fprint(2, "	loc %S\n", href);
		newpath(s, smprint("%S", href), 0, 0, 0);
		return 0;
	}

	/* relative path, prefix it by the path to the page just read */
	if(Dumpparse)
		fprint(2, "	rel %s/%S\n", path, href);
	newpath(s, smprint("%s/%S", path, href), 0, 0, 0);

	return 0;
}

int
parsebuf(Sess *s, char *path, Growbuf *gb)
{
	Iimage *i;
	Anchor *a;
	Docinfo *di;
	Item *items;
	Rune *rsrc;

	rsrc = runesmprint("%s/%s", s->url, path);
	if((items = parsehtml((uchar *)gb->buf, gb->len, rsrc, TextHtml, charset(gb->buf), &di)) == nil){
		free(rsrc);
		return 0;
	}

	for(a = di->anchors; a; a = a->next)
		resolve(s, di->base, a->href, path);
	for(i = di->images; i; i = i->nextimage)
		resolve(s, di->base, i->imsrc, path);

	freeitems(items);
	freedocinfo(di);
	free(rsrc);
	return 1;
}

int
parse(Sess *s, char *path)
{
	Growbuf *gb;
	int n, rc, fd;
	char tmp[1024];
	char **t, *tags[] = {
		"<html", "<meta", "<!doctype", "<link", "<title",
		"<script", "<body", "<header", "<li>", "<!--", nil
	};

	if((fd = exch(s, "GET", nil, path)) == -1)
		return -1;								/* does not exist */

	if((n = read(fd, tmp, sizeof(tmp)-1)) > 0){
		tmp[sizeof(tmp)-1] = 0;
		for(t = tags; *t; t++)
			if(cistrstr(tmp, *t) == nil)
				break;
		if(*t == nil){
fprint(2, "%s - not http %s\n", path, tmp);
			close(fd);
			return 0;							/* exists but not html */
		}
	}

	gb = allocbuf();
	do{
		growbuf(gb, tmp, n);
	}while((n = read(fd, tmp, sizeof(tmp)-1)) > 0);
	close(fd);

	if(gb->buf == nil || *gb->buf == 0){
		werrstr("empty file");
		freebuf(gb);
		return 0;								/* exists but not html */
	}
	rc = parsebuf(s, path, gb);
	freebuf(gb);
	return rc;
}

